###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         peekabooyar.py                                                      #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import traceback
from peekaboo.ruleset import Result, RuleResult
import yara


def contains_peekabooyar(config, s):
    """
    Checks the given sample for the PeekabooYar (EICAR like) malicious string.

    :param s: sample to check
    :return: RueleResult
    """
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    rules = yara.compile(
        source='''
        rule peekabooyar
        {
             strings:
                  $peekabooyar1 = "X5O!P%@AP-/_(:)_/-X22x8cz2$PeekabooAV-STD-ANTIVIRUS-TEST-FILE!$H+H*"
        
             condition:
                  $peekabooyar1
        }'''
    )

    with open(s.get_file_path(), 'rb') as f:
        matches = rules.match(data=f.read())

    if matches != []:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Die Datei beinhaltet Peekabooyar.",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Die Datei beinhaltet kein erkennbares Peekabooyar.",
                      further_analysis=True)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# queuing.py                                                                  #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
from threading import Thread
from Queue import Queue
from peekaboo import Singleton
from peekaboo.ruleset.engine import RulesetEngine
from peekaboo.exceptions import CuckooReportPendingException


logger = logging.getLogger(__name__)


def create_workers(worker_count=4):
    """
    Create n Peekaboo worker threads to process samples.

    :param worker_count: The amount of worker threads to create. Defaults to 4.
    """
    for i in range(0, worker_count):
        logger.debug("Create Worker %d" % i)
        w = Worker(i)
        JobQueue.workers.append(w)
        w.start()
    logger.info('Created %d Workers.' % worker_count)


class JobQueue(Singleton):
    """
    Peekaboo's queuing system.

    @author: Sebastian Deiss
    """
    workers = []
    jobs = Queue()

    @staticmethod
    def submit(sample, submitter, timeout=300):
        """
        Adds a Sample object to the job queue.
        If the queue is full, we block for 300 seconds and then throw an exception.

        :param sample: The Sample object to add to the queue.
        :param submitter: The name of the class / module that wants to submit the sample.
        :param timeout: Block until timeout is reached and then trow an exception
                        if the job has not been submitted.
        :raises Full: if the queue is full.
        """
        logger.debug("New sample submitted to job queue by %s. %s" % (submitter, sample))
        # thread safe most likely no race condition possible
        JobQueue.jobs.put(sample, True, timeout)


class Worker(Thread):
    """
    A Worker thread to process a sample.

    @author: Sebastian Deiss
    """
    def __init__(self, wid):
        self.active = True
        self.worker_id = wid
        Thread.__init__(self)

    def run(self):
        while self.active:
            logger.debug('Worker is ready')
            sample = JobQueue.jobs.get(True)  # wait blocking for next job (thread safe)
            logger.info('Worker %d: Processing sample %s' % (self.worker_id, sample))

            sample.init()

            try:
                engine = RulesetEngine(sample)
                engine.run()
                engine.report()
            except CuckooReportPendingException:
                pass
            except Exception as e:
                logger.exception(e)

    def __exit__(self, exc_type, exc_value, traceback):
        self.active = False



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# __init__.py                                                                 #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import re


VERSION = (1, 6)
AUTHORS = ['Felix Bauer', 'Sebastian Deiss']

__version__ = '.'.join(map(str, VERSION))
__author__ = ', '.join(AUTHORS)
__description__ = 'Peekaboo Extended Email Attachment Behavior Observation Owl'
__copyright__ = 'Copyright (C) 2016-2018 science + computing ag. All rights reserved.'
__license__ = 'GPLv3'


_owl = """
PEEKABOO {0}

Peekaboo Extended Email Attachment Behavior Observation Owl

                   _a_aa                    a_aa,
                    '*U4UUUULa_aa_aa_aajUUU4XU7'
                      aX''''''UUXU4XUU'''''!Ua
                    _U'        -U4UU'   _    'U,
                    ?i   jLd1   ?#Wi   4L01   Ui
                    -U,        4#000P        _U'
                     -*Xa_a_a_WUW##KUL_a_a_aX7'
                    _aXUXUUU4UUX4XX444UUUUUUXLa,
                   _UXXUXUXU47'!'!'!'!*X444U4UXX,
                   ?XU4U4''   _   __   -'UUXUUi
                   ?4U4'     / | / /_     'UUXi
                    *Xi      | || '_ \     ?X7
                     *L      | || (_) |     j7
                      *a     |_(_)___/      jY
                       -L,                _/'
                         'l,            _/'
                           j7_a_;  aaa/4
               _aaaaaa#0000#00000##0##00000000aaaaaa,
        aaad0P!!!!!!                             '!!!!!!Laaa
  _aa!!!!                                                    !! _,
(never mind the K)
""".format(__version__)


#
# Helpers
#


class MultiRegexMatcher(object):
    """
    Validate multiple regular expressions for the same string.

    @author: Sebastian Deiss
    """
    def __init__(self, patterns, flags=0):
        self.__patterns = [(re.compile(pattern, flags)) for pattern in patterns]
        self.matched_pattern = -1    # No pattern matched (default value)

    def match(self, str):
        """
        Try to apply the patterns at the start of the string.
        When the first pattern that matches, processing is stopped
        and a match object is returned.

        :param str: The string to apply the pattern to.
        :return: a match object, or None if no match was found.
        """
        iter_count = 0
        for pattern in self.__patterns:
            _match = re.match(pattern, str)
            if _match:
                self.matched_pattern = iter_count
                return _match
            iter_count += 1
        return None


class Singleton(type):
    """
    A Singleton in Python.

    @see: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# config.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import sys
import logging
from ConfigParser import SafeConfigParser, NoSectionError, NoOptionError


logger = logging.getLogger(__name__)
_config = None


def parse_config(config_file):
    """
    Parse the Peekaboo configuration file.

    :param config_file: Path to the configuration file.
    :return: A PeekabooConfig object containing the configuration options.
    """
    global _config
    if _config is None:
        _config = PeekabooConfig(config_file)
    return _config


def _set_config(config):
    """
    Set the configuration object manually. This is required e. g. for unit tests
    with a mocked configuration object.

    :param config: The mocked configuration object.
    :return: The given configuration object.
    """
    global _config
    if _config is None:
        _config = config
    return _config


def get_config():
    """ Get the Peekaboo configuration object. """
    assert _config is not None
    return _config


class PeekabooConfig(object):
    """
    This class represents the Peekaboo configuration file.

    :author: Sebastian Deiss
    """
    def __init__(self, config_file='./peekaboo.conf'):
        self.__config = None
        self.user = None
        self.group = None
        self.pid_file = None
        self.sock_file = None
        self.log_level = logging.INFO
        self.log_format = '%(asctime)s - %(name)s - (%(threadName)s) - ' \
                          '%(levelname)s - %(message)s'
        self.interpreter = None
        self.chown2me_exec = None
        self.worker_count = 3
        self.sample_base_dir = None
        self.job_hash_regex = None
        self.use_debug_module = None
        self.keep_mail_data = None
        self.db_url = None
        self.ruleset_config = None
        self.cuckoo_storage = None
        self.cuckoo_exec = None
        self.cuckoo_submit = None
        self.db_con = None
        ##############################################
        # setup default logging to log any errors during the
        # parsing of the config file.
        self.__setup_logging()
        self.__parse(config_file)

    def __parse(self, config_file):
        config = SafeConfigParser()
        config.read(config_file)
        self.__config = config
        try:
            log_level = config.get('logging', 'log_level')
            self.log_level = self.__parse_log_level(log_level)
            self.log_format = config.get('logging', 'log_format')
            self.user = config.get('global', 'user')
            self.group = config.get('global', 'group')
            self.pid_file = config.get('global', 'pid_file')
            self.sock_file = config.get('global', 'socket_file')
            self.interpreter = config.get('global', 'interpreter')
            self.chown2me_exec = config.get('global', 'chown2me_exec')
            self.worker_count = int(config.get('global', 'worker_count'))
            self.sample_base_dir = config.get('global', 'sample_base_dir')
            self.job_hash_regex = config.get('global', 'job_hash_regex')
            self.use_debug_module = True if config.get(
                'global', 'use_debug_module'
            ) == 'yes' else False
            self.keep_mail_data = True if config.get(
                'global', 'keep_mail_data'
            ) == 'yes' else False
            self.db_url = config.get('db', 'url')
            self.ruleset_config = config.get('ruleset', 'config')
            self.cuckoo_storage = config.get('cuckoo', 'storage_path')
            self.cuckoo_exec = config.get('cuckoo', 'exec')
            self.cuckoo_submit = config.get('cuckoo', 'submit').split(' ')
            # Update logging with what we just parsed from the config
            self.__setup_logging()
        except NoSectionError as e:
            logger.critical('configuration section not found')
            logger.exception(e)
            sys.exit(1)
        except NoOptionError as e:
            logger.critical('configuration option not found')
            logger.exception(e)
            sys.exit(1)

    def change_log_level(self, log_level):
        """
        Overwrite the log level from the configuration file.

        :param log_level: The new log level.
        """
        ll = self.__parse_log_level(log_level)
        self.log_level = ll
        logger.setLevel(ll)

    def add_db_con(self, db_con):
        self.db_con = db_con

    def get_db_con(self):
        if self.db_con:
            return self.db_con
        raise ValueError('Database connection is not configured.')

    def __parse_log_level(self, log_level):
        if log_level == 'CRITICAL':
            return logging.CRITICAL
        elif log_level == 'ERROR':
            return logging.ERROR
        elif log_level == 'WARNING':
            return logging.WARNING
        elif log_level == 'INFO':
            return logging.INFO
        elif log_level == 'DEBUG':
            return logging.DEBUG

    def __setup_logging(self):
        """
        Setup logging to console.
        """
        _logger = logging.getLogger()

        # Check if we already have a log handler
        if len(_logger.handlers) > 0:
            # Remove all handlers
            for handler in _logger.handlers:
                _logger.removeHandler(handler)
        # log format
        log_formatter = logging.Formatter(self.log_format)
        # create console handler and set level to debug
        to_console_log_handler = logging.StreamHandler(sys.stdout)
        to_console_log_handler.setFormatter(log_formatter)
        _logger.addHandler(to_console_log_handler)
        _logger.setLevel(self.log_level)

    def __str__(self):
        sections = {}
        for section in self.__config.sections():
            sections[section] = {}
            for key, value in self.__config.items(section):
                sections[section][key] = value
        return '<PeekabooConfig(%s)>' % str(sections)

    __repr__ = __str__


class PeekabooRulesetConfiguration(object):
    """
    This class represents the ruleset configuration file "ruleset.conf".

    The ruleset configuration is stored as a dictionary in the form of
    ruleset_config[rule_name][config_option] = value | [value1, value2, ...]

    @author: Sebastian Deiss
    @since: 1.6
    """
    def __init__(self, config_file):
        self.config_file = config_file
        self.ruleset_config = {}

    def parse(self):
        config = SafeConfigParser()
        try:
            config.read(self.config_file)
            for section in config.sections():
                if section not in self.ruleset_config.keys():
                    self.ruleset_config[section] = {}
                for setting, value in config.items(section):
                    if '.' in setting:
                        key = setting.split('.')[0]
                        if key not in self.ruleset_config[section]:
                            self.ruleset_config[section][key] = []
                        self.ruleset_config[section][key].append(value)
                    else:
                        self.ruleset_config[section][setting] = value
        except NoSectionError as e:
            logger.exception(e)
        except NoOptionError as e:
            logger.exception(e)

    def get_config(self):
        return self.ruleset_config

    def __str__(self):
        return '<PeekabooRulesetConfiguration(filepath="%s")>' % self.config_file

    __repr__ = __str__



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         cuckoo.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import re
import os
import logging
import json
import subprocess
from twisted.internet import protocol
from peekaboo import MultiRegexMatcher
from peekaboo.config import get_config
from peekaboo.exceptions import CuckooAnalysisFailedException
from peekaboo.toolbox.sampletools import ConnectionMap
from peekaboo.queuing import JobQueue


logger = logging.getLogger(__name__)


def submit_to_cuckoo(sample):
    """
    Submit a file or directory to Cuckoo for behavioural analysis.

    :param sample: Path to a file or a directory.
    :return: The job ID used by Cuckoo to identify this analysis task.
    """
    config = get_config()
    try:
        proc = config.cuckoo_submit
        proc.append(sample)
        p = subprocess.Popen(proc,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        p.wait()
    except Exception as e:
        raise CuckooAnalysisFailedException(e)

    if not p.returncode == 0:
        # TODO: tell opponent on socket that file has not been checked.
        raise CuckooAnalysisFailedException('cuckoo submit returned a non-zero return code.')
    else:
        out, err = p.communicate()
        logger.debug("cuckoo submit STDOUT: %s" % out)
        logger.debug("cuckoo submit STDERR: %s" % err)
        # process output to get job ID
        patterns = list()
        # Example: Success: File "/var/lib/peekaboo/.bashrc" added as task with ID #4
        patterns.append(".*Success.*: File .* added as task with ID #([0-9]*).*")
        patterns.append(".*added as task with ID ([0-9]*).*")
        matcher = MultiRegexMatcher(patterns)
        response = out.replace("\n", "")
        m = matcher.match(response)
        logger.debug('Pattern %d matched.' % matcher.matched_pattern)

        if m:
            job_id = int(m.group(1))
            return job_id
        raise CuckooAnalysisFailedException(
            'Unable to extract job ID from given string %s' % response
        )


class CuckooServer(protocol.ProcessProtocol):
    """
    Class that is used by twisted.internet.reactor to process Cuckoo
    output and process its behavior.

    Usage:
    srv = CuckooServer()
    reactor.spawnProcess(srv, 'python2', ['python2', '/path/to/cukoo.py'])
    reactor.run()

    @author: Felix Bauer
    @author: Sebastian Deiss
    """
    def __init__(self):
        self.__report = None

    def connectionMade(self):
        logger.info('Connected. Cuckoo PID: %s' % self.transport.pid)
        return None

    def outReceived(self, data):
        """ on receiving output on STDOUT from Cuckoo """
        logger.debug('STDOUT %s' % str(data))

    def errReceived(self, data):
        """ on receiving output on STDERR from Cuckoo """
        logger.debug('STDERR %s' % str(data.replace('\n', '')))

        #
        # FILE SUBMITTED
        # printed out but has no further effect
        #
        # 2016-04-12 09:14:06,984 [lib.cuckoo.core.scheduler] INFO: Starting
        # analysis of FILE "cuckoo.png" (task #201, options "")
        # INFO: Starting analysis of FILE ".bashrc" (task #4, options "")
        m = re.match('.*INFO: Starting analysis of FILE \"(.*)\" \(task #([0-9]*), options .*', data)

        if m:
            logger.info("File submitted: task #%s, filename %s" % (m.group(2),
                                                                   m.group(1)))

        #
        # ANALYSIS DONE
        #
        # 2016-04-12 09:25:27,824 [lib.cuckoo.core.scheduler] INFO: Task #202:
        # reports generation completed ...
        m = re.match(".*INFO: Task #([0-9]*): reports generation completed.*",
                     data)
        if m:
            job_id = int(m.group(1))
            logger.debug("Analysis done for task #%d" % job_id)
            logger.debug("Remaining connections: %d" % ConnectionMap.size())
            sample = ConnectionMap.get_sample_by_job_id(job_id)
            if sample:
                logger.debug('Requesting Cuckoo report for sample %s' % sample)
                self.__report = CuckooReport(job_id)
                sample.set_attr('cuckoo_report', self.__report)
                sample.set_attr('cuckoo_json_report_file', self.__report.file_path)
                JobQueue.submit(sample, self.__class__)
                logger.debug("Remaining connections: %d" % ConnectionMap.size())
            else:
                logger.debug('No connection found for ID %d' % job_id)

    def inConnectionLost(self):
        logger.debug("Cuckoo closed STDIN")
        os._exit(1)

    def outConnectionLost(self):
        logger.debug("Cuckoo closed STDOUT")
        os._exit(1)

    def errConnectionLost(self):
        logger.warning("Cuckoo closed STDERR")
        os._exit(1)

    def processExited(self, reason):
        logger.info("Cuckoo exited with status %d" % reason.value.exitCode)
        os._exit(0)

    def processEnded(self, reason):
        logger.info("Cuckoo ended with status %d" % reason.value.exitCode)
        os._exit(0)


class CuckooReport(object):
    """
    Represents a Cuckoo analysis JSON report.

    @author: Sebastian Deiss
    """
    def __init__(self, job_id):
        self.job_id = job_id
        self.file_path = None
        self.report = None
        self._parse()
        self.errors = self.report['debug']['errors']

    def _parse(self):
        """
        Reads the JSON report from Cuckoo and loads it into the Sample object.
        """
        config = get_config()
        cuckoo_report = os.path.join(
            config.cuckoo_storage, 'analyses/%d/reports/report.json'
                                   % self.job_id
        )

        if not os.path.isfile(cuckoo_report):
            raise OSError('Cuckoo report not found at %s.' % cuckoo_report)
        else:
            logger.debug(
                'Accessing Cuckoo report for task %d at %s '
                % (self.job_id, cuckoo_report)
            )
            self.file_path = cuckoo_report
            with open(cuckoo_report) as data:
                try:
                    report = json.load(data)
                    self.report = report
                except ValueError as e:
                    logger.exception(e)

    @property
    def requested_domains(self):
        """
        Gets the requested domains from the Cuckoo report.

        :return: The requested domains from the Cuckoo report.
        """
        try:
            return [d['request'] for d in self.report['network']['dns']]
        except KeyError:
            return None

    @property
    def signatures(self):
        """
        Gets the triggered signatures from the Cuckoo report.

        :return: The triggered signatures from the Cuckoo report.
        """
        return self.report['signatures']

    @property
    def score(self):
        """
        Gets the score from the Cuckoo report.

        :return: The score from the Cuckoo report.
        """
        return self.report['info']['score']

    @property
    def analysis_failed(self):
        """
        Has the Cuckoo analysis failed?

        :return: True if the Cuckoo analysis failed, otherwise False.
        """
        if self.errors:
            logger.warning('Cuckoo run_analysis failed. Reason: %s' % str(self.errors))
            return True
        return False



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# exceptions.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


class PeekabooException(Exception):
    """ General exception class for all custom exception classes of Peekaboo. """
    pass


class PeekabooConfigException(PeekabooException):
    pass


class PeekabooDatabaseError(PeekabooException):
    pass


class PeekabooRulesetException(PeekabooException):
    pass


class CuckooReportPendingException(PeekabooRulesetException):
    pass


class CuckooAnalysisFailedException(PeekabooException):
    pass



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         sampletools.py                                                      #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
import string
import threading
from random import choice
from datetime import datetime
from ConfigParser import SafeConfigParser
from peekaboo import Singleton
from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)


class SampleMetaInfo(object):
    """
    Additional meta information about a Sample.

    @author: Felix Bauer
    @author: Sebastian Deiss
    """
    def __init__(self, meta_info_file):
        self.__meta_info_file = meta_info_file
        self.meta_info = None
        self._parse()

    def _parse(self):
        """
        Parse a meta information file.

        @see: SafeConfigParser
        """
        logger.debug('Parsing sample metadata from %s' % self.__meta_info_file)
        self.meta_info = SafeConfigParser()
        self.meta_info.read(self.__meta_info_file)

    def get_all(self):
        """
        Gets the parsed meta info file.

        :return: A ConfigParser instance for the parsed meta info file.
        """
        return self.meta_info

    def get_mime_type(self):
        """
        Gets the MIME type parsed from the meta info file field 'type_declared'.

        :return: The MIME type from the meta info file field 'type_declared'.
        """
        return self.meta_info.get('attachment', 'type_declared')

    def __str__(self):
        return '<SampleMetaInfo(%s)>' % str(self.meta_info)


class ConnectionMap(Singleton):
    """
    Maps socket objects with one or more samples.
    This is required for the reporting, so we know which
    Sample objects belong to which socket connection.

    @author: Sebastian Deiss
    """
    __lock = threading.RLock()
    __map = {}

    @staticmethod
    def add(socket, sample):
        """
        Add an entry to the connection map. An entry consists of a
        socket object and a list of Sample objects.

        :param socket: The socket object to add.
        :param sample: The corresponding Samples objects for the socket.
        :return:The length of the connection map.
        """
        with ConnectionMap.__lock:
            logger.debug('Registered sample for connection %s' % socket)
            if ConnectionMap.has_connection(socket):
                ConnectionMap.__map[socket].append(sample)
            else:
                ConnectionMap.__map[socket] = [sample]
            return ConnectionMap.size()

    @staticmethod
    def remove(socket, sample):
        """
        Remove a Sample or an entry from the connection map. First, we
        remove the given Sample object from the list of Sample objects of
        the given socket. If the sample list is empty, we remove the
        entire entry from the map.

        :param socket: A socket object, which is related to the sample.
        :param sample: The sample to remove.
        :return: The length of the connection map.
        """
        with ConnectionMap.__lock:
            if ConnectionMap.has_connection(socket):
                logger.debug(
                    'Removing sample for connection %s, Sample: %s' % (socket, sample)
                )
                ConnectionMap.__map[socket].remove(sample)
                if len(ConnectionMap.__map[socket]) == 0:
                    ConnectionMap.__map.pop(socket)
                    logger.debug('Removing connection: %s' % socket)
            else:
                logger.debug(
                    'Connection does not exist.'
                    'Connection: %s, Sample: %s, Map: %s'
                    % (socket, sample, ConnectionMap.__map)
                )
            return ConnectionMap.size()

    @staticmethod
    def size():
        """ Gets the length of the connection map. """
        return len(ConnectionMap.__map)

    @staticmethod
    def _dump():
        """ Get the connection map. This method might be useful for debugging. """
        return ConnectionMap.__map

    @staticmethod
    def has_connection(socket):
        """
        Check if the given socket object exists in the map.

        :param socket: A socket object to search for in the map.
        :return: True if the map contains the given socket object, otherwise False.
        """
        if socket in ConnectionMap.__map.keys():
            return True
        return False

    @staticmethod
    def get_sample_by_job_id(job_id):
        """
        Get a Sample object from the map by its job ID.

        :param job_id: The job ID of the Sample object to fetch.
        :return:The Sample object with the given job ID or None.
        """
        with ConnectionMap.__lock:
            logger.debug("Searching for a sample with job ID %d" % job_id)
            for __, samples in ConnectionMap.__map.iteritems():
                logger.debug('Samples for this connection: %s' % samples)
                for sample in samples:
                    if job_id == sample.job_id:
                        logger.debug('Found %s for job ID %d' % (sample, job_id))
                        return sample

    @staticmethod
    def get_sample_by_sha256(sha256sum):
        """
        Get a Sample object from the map by its SHA-256 checksum.

        :param sha256sum: The SHA-256 checksum of the file represented by a Sample object.
        :return: The Sample object with the given SHA-256 checksum or None.
        """
        with ConnectionMap.__lock:
            logger.debug(
                'Searching for a sample with SHA-256 checksum %s' % sha256sum
            )
            for __, samples in ConnectionMap.__map.iteritems():
                logger.debug('Samples for this connection: %s' % samples)
                for sample in samples:
                    if sha256sum == sample.sha256sum:
                        logger.debug(
                            'Found %s for SHA-256 hash %s' % (sample, sha256sum)
                        )
                        return sample

    @staticmethod
    def get_samples_by_sha256(sha256sum):
        """
        Get all Sample objects from the map with the same SHA-256 checksum.

        :param sha256sum: A SHA-256 checksum to search for.
        :return: The Sample objects with the given SHA-256 checksum or None.
        """
        with ConnectionMap.__lock:
            logger.debug('Searching for all samples with SHA-256 checksum %s' % sha256sum)
            matching_samples = []
            for __, samples in ConnectionMap.__map.iteritems():
                logger.debug('Samples for this connection: %s' % samples)
                for sample in samples:
                    if sha256sum == sample.sha256sum:
                        logger.debug('Found %s for SHA-256 hash %s' % (sample, sha256sum))
                        matching_samples.append(sample)
            return matching_samples

    @staticmethod
    def get_samples_by_connection(socket):
        """
        Get all Sample objects for a specific socket connection.

        :param socket: The socket object to get all Samples objects for.
        :return: A list of all Sample objects for the given socket.
        """
        with ConnectionMap.__lock:
            matching_samples = []
            for sock in ConnectionMap.__map.iteritems():
                if sock == socket:
                    matching_samples = ConnectionMap.__map[sock]
            return matching_samples

    @staticmethod
    def in_progress(sha256sum):
        """
        Check if a Sample object in the map has the state 'inProgress'.

        :param sha256sum: The SHA-256 checksum of the file represented by a Sample object.
        :return: True if the Sample object with the given SHA-256 checksum is
                 'inProgress' otherwise False.
        """
        sample = ConnectionMap.get_sample_by_sha256(sha256sum)
        if sample is not None and sample.get_result() == Result.inProgress:
            return True
        return False


def next_job_hash(size=8):
    """
    Generates a job hash (default: 8 characters).

    :param size The amount of random characters to use for a job hash.
                Defaults to 8.
    :return Returns a job hash consisting of a static prefix, a timestamp
            representing the time when the method was invoked, and random characters.
    """
    job_hash = 'peekaboo-run_analysis-'
    job_hash += '%s-' % datetime.now().strftime('%Y%m%dT%H%M%S')
    job_hash += ''.join(
        choice(string.digits + string.ascii_lowercase + string.ascii_uppercase)
        for _ in range(size)
    )
    return job_hash



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         engine.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import os
import logging
from shutil import copyfile
from peekaboo.config import PeekabooRulesetConfiguration, get_config
from peekaboo.ruleset import Result, RuleResult
from peekaboo.ruleset.rules import *
from peekaboo.toolbox.peekabooyar import contains_peekabooyar
from peekaboo.exceptions import CuckooReportPendingException
from peekaboo.toolbox.plugins.oneanalysis import OneAnalysis


logger = logging.getLogger(__name__)


class RulesetEngine(object):
    """
    Peekaboo's ruleset engine.

    @author: Sebastian Deiss
    @since: 1.6
    """
    rules = [
        known,
        file_larger_than,
        file_type_on_whitelist,
        file_type_on_greylist,
        cuckoo_evil_sig,
        cuckoo_score,
        office_macro,
        requests_evil_domain,
        cuckoo_analysis_failed,
        contains_peekabooyar,
        final_rule
    ]

    def __init__(self, sample):
        peekaboo_config = get_config()
        ruleset_config = PeekabooRulesetConfiguration(peekaboo_config.ruleset_config)
        ruleset_config.parse()
        self.sample = sample
        self.config = ruleset_config.get_config()
        self.one_analysis_tool = OneAnalysis()

    def run(self):
        # TODO: Integrate this special rule in the base ruleset.
        result = self.__exec_rule(
            self.config,
            self.sample,
            self.one_analysis_tool.already_in_progress
        )
        if not result.further_analysis:
            return

        for rule in RulesetEngine.rules:
            result = self.__exec_rule(self.config, self.sample, rule)
            if not result.further_analysis:
                return

        logger.info("Rules evaluated")

    def report(self):
        # TODO: might be better to do this for each rule individually
        self.sample.report()
        if self.sample.get_result() == Result.bad:
            dump_processing_info(self.sample)
        self.sample.save_result()
        # TODO: might be better to move this call to a separate function, since it's not related to reporting.
        self.one_analysis_tool.queue_identical_samples(self.sample)  # depends on already_in_progress

    def __exec_rule(self, config, sample, rule_function):
        """
        rule wrapper for in/out logging and reporting
        """
        rule_name = rule_function.func_name
        logger.debug("Processing rule '%s' for %s" % (rule_name, sample))

        try:
            # skip disabled rules.
            if rule_name in config.keys() and \
               'enabled' in config[rule_name].keys() and \
               config[rule_name]['enabled'] == 'no':
                logger.debug("Rule '%s' is disabled." % rule_name)
                result = RuleResult(rule_name, result=Result.unchecked,
                                    reason="Regel '%s' ist deaktiviert." % rule_name,
                                    further_analysis=True)
            else:
                result = rule_function(config, sample)
            sample.add_rule_result(result)
        except CuckooReportPendingException as e:
            # in case the Sample is requesting the Cuckoo report
            raise
        # catch all other exceptions for this rule
        except Exception as e:
            logger.warning("Unexpected error in '%s' for %s" % (rule_name,
                                                                sample))
            logger.exception(e)
            # create "fake" RuleResult
            result = RuleResult("RulesetEngine", result=Result.unknown,
                                reason="Regel mit Fehler abgebrochen",
                                further_analysis=True)
            sample.add_rule_result(result)

        logger.info("Rule '%s' processed for %s" % (rule_name, sample))
        return result


def dump_processing_info(sample):
    """
    Saves the Cuckoo report as HTML + JSON and the meta info file (if available)
    to a directory named after the job hash.
    """
    job_hash = sample.get_job_hash()
    dump_dir = os.path.join(os.environ['HOME'], 'malware_reports', job_hash)
    if not os.path.isdir(dump_dir):
        os.makedirs(dump_dir, 0770)
    filename = sample.get_filename() + '-' + sample.sha256sum

    logger.debug('Dumping processing info to %s for sample %s' % (dump_dir, sample))

    # meta info file
    if sample.has_attr('meta_info_file'):
        try:
            copyfile(sample.get_attr('meta_info_file'),
                     os.path.join(dump_dir, filename + '.info'))
        except Exception as e:
            logger.exception(e)

    # Peekaboo's report
    try:
        with open(os.path.join(dump_dir, filename + '_report.txt'), 'w+') as f:
            f.write(sample.get_peekaboo_report())
    except Exception as e:
        logger.exception(e)

    if sample.has_attr('cuckoo_json_report_file'):
        # Cuckoo report
        try:
            # JSON
            copyfile(sample.get_attr('cuckoo_json_report_file'),
                     os.path.join(dump_dir, filename + '.json'))
        except Exception as e:
            logger.exception(e)
        try:
            # HTML
            copyfile(sample.get_attr('cuckoo_json_report_file').replace('json', 'html'),
                     os.path.join(dump_dir, filename + '.html'))
        except Exception as e:
            logger.exception(e)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         rules.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import traceback
import re
import logging
from peekaboo.config import get_config
from peekaboo.ruleset import Result, RuleResult


logger = logging.getLogger(__name__)


def known(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    db = get_config().get_db_con()
    if db.known(s):
        sample_info = db.sample_info_fetch(s)
        return RuleResult(position,
                          result=sample_info.get_result(),
                          reason=sample_info.reason,
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei ist dem System noch nicht bekannt",
                      further_analysis=True)


def file_larger_than(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    size = int(config['file_larger_than']['bytes'])
    if s.file_size > size:
        return RuleResult(position,
                          result=Result.unknown,
                          reason="Datei hat mehr als %d bytes"
                          % size,
                          further_analysis=True)

    return RuleResult(position,
                      result=Result.ignored,
                      reason="Datei ist nur %d bytes lang"
                      % s.file_size,
                      further_analysis=False)


def file_type_on_whitelist(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    whitelist = config['file_type_on_whitelist']['whitelist']

    if s.mimetype in whitelist:
        return RuleResult(position,
                          result=Result.ignored,
                          reason="Dateityp ist auf Whitelist",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Dateityp ist nicht auf Whitelist",
                      further_analysis=True)


def file_type_on_greylist(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    greylist = config['file_type_on_greylist']['greylist']

    if s.mimetype in greylist:
        return RuleResult(position,
                          result=Result.unknown,
                          reason="Dateityp ist auf der Liste der zu analysiserenden Typen",
                          further_analysis=True)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Dateityp ist nicht auf der Liste der zu analysierenden Typen (%s)"
                      % (str(s.mimetypes)),
                      further_analysis=False)


def cuckoo_evil_sig(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    # signatures that if matched mark a sample as bad
    # list all installed signatures
    # grep -o "description.*" -R . ~/cuckoo2.0/modules/signatures/
    bad_sigs = config['cuckoo_evil_sig']['signature']

    sigs = []

    # look through matched signatures
    for descr in s.cuckoo_report.signatures:
        logger.debug(descr['description'])
        sigs.append(descr['description'])

    # check if there is a "bad" signatures and return bad
    matched_bad_sigs = []
    for sig in bad_sigs:
        match = re.search(sig, str(sigs))
        if match:
            matched_bad_sigs.append(sig)

    if len(matched_bad_sigs) == 0:
        return RuleResult(position,
                          result=Result.unknown,
                          reason="Keine Signatur erkannt die auf Schadcode hindeutet",
                          further_analysis=True)

    return RuleResult(position,
                      result=Result.bad,
                      reason="Folgende Signaturen wurden erkannt: %s"
                      % ''.ljust(8).join(["%s\n" % s for s in matched_bad_sigs]),
                      further_analysis=False)


def cuckoo_score(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    threshold = float(config['cuckoo_score']['higher_than'])
    if s.cuckoo_report.score >= threshold:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Cuckoo score >= %s: %s"
                          % (threshold, s.cuckoo_report.score),
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Cuckoo score < %s: %s"
                      % (threshold, s.cuckoo_report.score),
                      further_analysis=True)


def office_macro(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    if s.office_macros:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Die Datei beinhaltet ein Office-Makro",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Die Datei beinhaltet kein erkennbares Office-Makro",
                      further_analysis=True)


def requests_evil_domain(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    evil_domains = config['requests_evil_domain']['domain']

    for d in s.requested_domains:
        if d in evil_domains:
            return RuleResult(position,
                              result=Result.bad,
                              reason="Die Datei versucht mindestens eine Domain aus der Blacklist zu kontaktieren (%s)" % d,
                              further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei scheint keine Domains aus der Blacklist kontaktieren zu wollen",
                      further_analysis=True)


def cuckoo_analysis_failed(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    if s.cuckoo_analysis_failed:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Die Verhaltensanalyse durch Cuckoo hat einen Fehler Produziert und konnte nicht erfolgreich abgeschlossen werden",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Die Verhaltensanalyse durch Cuckoo wurde erfolgreich abgeschlossen",
                      further_analysis=True)


def final_rule(config, s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei scheint keine erkennbaren Schadroutinen zu starten",
                      further_analysis=True)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         ms_office.py                                                        #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
from oletools.olevba import VBA_Parser


logger = logging.getLogger(__name__)
ms_office_extensions = [
    ".doc", ".docm", ".dotm", ".docx",
    ".ppt", ".pptm", ".pptx", ".potm", ".ppam", ".ppsm",
    ".xls", ".xlsm", ".xlsx",
]


def has_office_macros(office_file):
    """
    Detects macros in Microsoft Office documents.

    :param office_file: The MS Office document to check for macros.
    :return: True if macros where found, otherwise False.
             If VBA_Parser crashes it returns False too.
    """
    file_extension = office_file.split('.')[-1]
    if file_extension not in ms_office_extensions:
        return False
    try:
        # VBA_Parser reports macros for office documents
        vbaparser = VBA_Parser(office_file)
        return vbaparser.detect_vba_macros()
    except TypeError:
        # The given file is not an office document.
        return False
    except Exception as e:
        logger.exception(e)
        return False



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# sample.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import os
import hashlib
import re
import errno
import shutil
import logging
from datetime import datetime
from peekaboo.config import get_config
from peekaboo.exceptions import CuckooReportPendingException, \
                                CuckooAnalysisFailedException
from peekaboo.toolbox.sampletools import SampleMetaInfo, ConnectionMap, next_job_hash
from peekaboo.toolbox.files import chown2me, guess_mime_type_from_file_contents
from peekaboo.toolbox.ms_office import has_office_macros
from peekaboo.toolbox.cuckoo import submit_to_cuckoo
import peekaboo.ruleset as ruleset


logger = logging.getLogger(__name__)


def make_sample(file, socket):
    """
    Create a Sample object from a given file.

    :param file: Path to the file to create a Sample object from.
    :param socket: An optional socket to write the report to.
    :return: A sample object representing the given file or None if the file does not exist.
    """
    logger.debug("Looking at file %s" % file)
    if not os.path.isfile(file):
        logger.debug('%s is not a file' % file)
        return None
    s = Sample(file, socket)
    logger.debug('Created sample %s' % s)
    return s


class Sample(object):
    """
    This class handles and describes samples to be analysed by Peekaboo.

    A sample has attributes like:
    filename, MIME type, sha256, ...
    Those attributes are determined on demand kept in a dictionary, which is
    accessible through the methods has_attr, get_attr, and set_attr.

    The data structure works together with Cuckoo to run behavioral attributes.

    @author: Felix Bauer
    @author: Sebastian Deiss
    """
    def __init__(self, file_path, sock=None):
        self.__path = file_path
        self.__config = get_config()
        self.__db_con = self.__config.get_db_con()
        self.__meta_info = None
        self.__wd = None
        self.__filename = os.path.basename(self.__path)
        # A symlink that points to the actual file named
        # sha256sum.suffix
        self.__symlink = None
        self.__result = ruleset.Result.unchecked
        self.__report = []  # Peekaboo's own report
        self.__socket = sock
        # Additional attributes for a sample object (e. g. meta info)
        self.__attributes = {}
        self.initialized = False
        self.meta_info_loaded = False

    def init(self):
        """
        Initialize the Sample object.

        The actual initialization is done here, because the main thread should
        not do the heavy lifting of e. g. parsing the meta info file to be able
        to accept new connections as quickly as possible.
        Instead, it only adds the sample objects to the queue and the workers
        to the actual initialization.
        """
        if self.initialized:
            return

        logger.debug("initializing sample")

        job_hash = self.get_job_hash()
        self.__wd = os.path.join(self.__config.sample_base_dir, job_hash)

        chown2me()

        meta_info_file = os.path.join(self.__wd, self.__filename + '.info')
        self.set_attr('meta_info_file', meta_info_file)
        self.load_meta_info(meta_info_file)

        try:
            self.__create_symlink()
        except OSError:
            pass
        self.initialized = True

        # Add sample to database with state 'inProgress' if the sample is unknown
        # to avoid multiple concurrent analysis.
        self.__result = ruleset.Result.inProgress
        self.__db_con.analysis2db(self)

        message = "Datei \"%s\" %s wird analysiert\n" % (self.__filename,
                                                         self.sha256sum)
        self.__report.append(message)
        self.__send_message(message)

    def get_attr(self, key):
        """
        Get a sample attribute by a specified key.

        :param key: The identifier of the sample attribute to get.
        """
        if self.has_attr(key):
            return self.__attributes[key]
        raise KeyError("Attribute for key '%s' not found." % key)

    def set_attr(self, key, val, override=True):
        """
        Add an attribute to a sample.

        :param key: The identifier of the attribute.
        :param val: The attribute to add.
        :param override: Whether the existing attribute shall be overwritten or not.
        """
        if self.has_attr(key) and override is False:
            raise KeyError("Key '%s' already exists." % key)
        self.__attributes[key] = val

    def has_attr(self, key):
        """
        Check if an attribute exists for this sample.

        :param key: The identifier of the attribute.
        """
        if key in self.__attributes.keys():
            return True
        return False

    def remove_attr(self, key):
        """
        Delete an attribute for this sample.

        :param key: The identifier of the attribute
        :raises ValueError if the given key was not found in
                the attributes dictionary.
        """
        if key in self.__attributes.keys():
            del self.__attributes[key]
        raise ValueError('No attribute named "%s" found.' % key)

    def get_file_path(self):
        return self.__path

    def get_filename(self):
        return self.__filename

    def get_result(self):
        return self.__result

    def get_peekaboo_report(self):
        return ''.join(self.__report)

    def get_job_hash(self):
        job_hash = re.sub(self.__config.job_hash_regex, r'\1',
                          self.__path)
        if job_hash == self.__path:
            # regex did not match.
            # so we generate our own job hash and create the
            # working directory.
            job_hash = next_job_hash()
            os.mkdir(os.path.join(self.__config.sample_base_dir,
                                  job_hash))

        logger.debug("Job hash for this sample: %s" % job_hash)
        return job_hash

    def load_meta_info(self, meta_info_file):
        try:
            self.__meta_info = SampleMetaInfo(meta_info_file)
            logger.debug('Parsing meta info file %s for file %s' % (meta_info_file, self.__path))
            # Add the information from the dump info file as attributes to the sample object.
            for info in self.__meta_info.get_all().items('attachment'):
                logger.debug('meta_info_%s = %s' % (info[0], info[1]))
                self.set_attr('meta_info_' + info[0], info[1])
            self.meta_info_loaded = True
        except Exception:
            logger.info('No metadata available for file %s' % self.__path)

    def save_result(self):
        if self.__db_con.known(self):
            logger.debug('Known sample info not logged to database')
        else:
            logger.debug('Saving results to database')
            self.__db_con.sample_info_update(self)
        if self.__socket is not None:
            ConnectionMap.remove(self.__socket, self)
        if not ConnectionMap.has_connection(self.__socket):
            self.__cleanup_temp_files()
            self.__close_socket()

    def add_rule_result(self, res):
        logger.debug('Adding rule result %s' % str(res))
        rule_results = []
        if self.has_attr('rule_results'):
            rule_results = self.get_attr('rule_results')
        rule_results.append(res)
        self.set_attr('rule_results', rule_results)

    def determine_result(self):
        for rule_result in self.get_attr('rule_results'):
            logger.debug("Current result: %s, Rule result: %s"
                         % (self.__result, rule_result.result))
            # check if result of this rule is worse than what we know so far
            if rule_result.result > self.__result:
                self.__result = rule_result.result
                self.set_attr('reason', rule_result.reason)

    def report(self):
        """
        Create the report for this sample. The report is saved as a list of
        strings and is available via get_peekaboo_report(). Also, if a socket connection was
        supplied to the sample the report messages are also written to the socket.
        """
        # TODO: move to rule processing engine.
        self.determine_result()

        for rule_result in self.get_attr('rule_results'):
            message = "Datei \"%s\": %s\n" % (self.__filename, str(rule_result))
            self.__report.append(message)
            self.__send_message(message)

        if self.__result == ruleset.Result.inProgress:
            logger.warning('Ruleset result forces to unchecked.')
            self.__result = ruleset.Result.unchecked

        message = "Die Datei \"%s\" wurde als \"%s\" eingestuft\n\n" \
                  % (self.__filename, self.__result.name)
        self.__report.append(message)
        self.__send_message(message)

    @property
    def sha256sum(self):
        if not self.has_attr('sha256sum'):
            with open(self.__path, 'rb') as f:
                checksum = hashlib.sha256(f.read()).hexdigest()
                self.set_attr('sha256sum', checksum)
                return checksum
        return self.get_attr('sha256sum')

    @property
    def known(self):
        _known = self.__db_con.known(self)
        if _known:
            self.set_attr('known', True)
            return True
        return False

    @property
    def file_extension(self):
        if self.has_attr('meta_info_name_declared'):
            file_ext = self.get_attr('meta_info_name_declared').split('.')[-1]
            if self.has_attr('file_extension'):
                if self.get_attr('file_extension') != file_ext:
                    self.set_attr('file_extension', file_ext, override=True)
            else:
                self.set_attr('file_extension', file_ext)
        elif not self.has_attr('file_extension'):
            file_ext = os.path.splitext(self.__filename)[1][1:]
            self.set_attr('file_extension', file_ext)
        return self.get_attr('file_extension')

    @property
    def mimetype(self):
        """
        Can not be cached (hard to determine if known/complete).

        determine mime on original p[0-9]* file
        later result will be "inode/symlink"
        """
        mime_type = None

        smime = {
            'p7s': [
                'application/pkcs7-signature',
                'application/x-pkcs7-signature',
                'application/pkcs7-mime',
                'application/x-pkcs7-mime',
            ]
        }

        # get MIME type from meta info
        try:
            declared_mt = self.__meta_info.get_mime_type()
            if declared_mt is not None:
                logger.debug('Sample declared as "%s"' % declared_mt)
                mime_type = declared_mt
        except Exception as e:
            logger.exception(e)
            if self.meta_info_loaded:
                logger.error('Cannot get mime type from meta info although meta info is loaded.')

        detected_mime_type = guess_mime_type_from_file_contents(self.__path)
        if detected_mime_type != mime_type:
            logger.debug(
                'Detected MIME type does not match declared MIME Type: declared: %s, detected: %s.'
                % (mime_type, detected_mime_type)
            )
            # check if the sample is an smime signature (smime.p7s)
            # If so, don't overwrite the MIME type since we do not want to analyse S/MIME signatures.
            try:
                declared_filename = self.get_attr('meta_info_name_declared')
            except KeyError:
                declared_filename = self.__filename
            if declared_filename == 'smime.p7s' and mime_type in smime['p7s']:
                logger.info('Using declared MIME type over detected one for S/MIME signatures.')
            else:
                logger.debug('Overwriting declared MIME Type with "%s"' % detected_mime_type)
                mime_type = detected_mime_type

        if not self.has_attr('mimetypes'):
            self.set_attr('mimetypes', mime_type)

        return self.get_attr('mimetypes')

    @property
    def job_id(self):
        if self.has_attr('job_id'):
            return self.get_attr('job_id')
        return -1

    @property
    def reason(self):
        # TODO: Cover all possible cases.
        # if reason exists and sample is not known?
        if not self.has_attr('reason'):
            if not self.has_attr('known'):
                rr = self.__db_con.fetch_rule_result(self)
                self.__result = rr.result
                self.set_attr('known', True)
                self.set_attr('reason',
                              'Ausschlaggebendes Ergebnis laut Datenbank: %s'
                              % rr.reason)
        return self.get_attr('reason')

    @property
    def office_macros(self):
        if not self.has_attr('office_macros'):
            self.set_attr('office_macros', has_office_macros(self.__path))
        return self.get_attr('office_macros')

    @property
    def file_size(self):
        if not self.has_attr('file_stat'):
            self.set_attr('file_stat', os.stat(self.__path))
        return self.get_attr('file_stat').st_size

    @property
    def analyses_time(self):
        if not self.has_attr('analyses_time'):
            timestamp = datetime.now().strftime("%Y%m%dT%H%M%S")
            self.set_attr('analyses_time', timestamp)
            return timestamp
        return self.get_attr('analyses_time')

    @property
    def requested_domains(self):
        if not self.has_attr('requested_domains'):
            try:
                self.set_attr(
                    'requested_domains',
                    self.get_attr('cuckoo_report').requested_domains
                )
            except KeyError:
                self.set_attr('requested_domains', [])
        return self.get_attr('requested_domains')

    @property
    def cuckoo_report(self):
        if not self.has_attr('cuckoo_report'):
            try:
                file_for_analysis = os.path.join(self.__wd, self.__symlink)
                logger.debug("Submitting %s to Cuckoo" % file_for_analysis)
                job_id = submit_to_cuckoo(file_for_analysis)
                self.set_attr('job_id', job_id)
                message = 'Erfolgreich an Cuckoo gegeben %s als Job %d\n' \
                          % (self, job_id)
                self.__report.append(message)
                logger.info('Sample submitted to Cuckoo. Job ID: %s. Sample: %s' % (job_id, self))
                self.__db_con.analysis_update(self)
                raise CuckooReportPendingException()
            except CuckooAnalysisFailedException as e:
                logger.exception(e)
        return self.get_attr('cuckoo_report')

    @property
    def cuckoo_analysis_failed(self):
        if not self.has_attr('cuckoo_failed'):
            if self.has_attr('cuckoo_report'):
                report = self.get_attr('cuckoo_report')
                if report.analysis_failed:
                    self.set_attr('cuckoo_failed', True)
                else:
                    self.set_attr('cuckoo_failed', False)
        return self.get_attr('cuckoo_failed')

    def __create_symlink(self):
        """
        creates a symlink to submit the file with the correct
        file extension to cuckoo via submit.py.
        """
        orig = os.path.join(self.__wd, self.__filename)
        self.__symlink = '%s/%s.%s' % (self.__wd,
                                       self.sha256sum,
                                       self.file_extension)
        logger.debug('ln -s %s %s' % (orig, self.__symlink))

        os.symlink(orig, self.__symlink)

    def __close_socket(self):
        logger.debug('Closing socket connection.')
        try:
            if self.__socket is not None:
                self.__socket.close()
        except EnvironmentError as e:
            # base class for exceptions that can occur outside the Python system.
            # e. g. IOError, OSError
            if e.errno == errno.EPIPE:
                logger.warning('Unable to close the socket. Broken pipe.')
            else:
                logger.exception(e)

    def __cleanup_temp_files(self):
        try:
            if self.__config.keep_mail_data:
                logger.debug('Keeping mail data in %s' % self.__wd)
            else:
                logger.debug("Deleting tempdir %s" % self.__wd)
                shutil.rmtree(self.__wd)
        except OSError as e:
            logger.exception(e)

    def __send_message(self, msg):
        """
        Write a message to the socket.

        :param msg: The message to send (max. 1024 bytes).
        """
        if self.__socket is None:
            return
        try:
            self.__socket.send(msg)
            logger.debug('Message send: %s ' % msg)
        except IOError as e:
            if e.errno == errno.EPIPE:
                logger.warning('Unable send message "%s". Broken pipe.' % msg)
            else:
                logger.exception(e)

    def __str__(self):
        meta_info_loaded = 'no'
        job_id = -1
        if self.__meta_info:
            meta_info_loaded = 'yes'
        if self.has_attr('job_id'):
            job_id = self.get_attr('job_id')

        return ("<Sample(filename='%s', known='%s', meta_info_loaded='%s', job_id='%d',"
                " result='%s', sha256sum='%s')>"
                % (self.__filename,
                   'yes' if self.known else 'no',
                   meta_info_loaded,
                   job_id,
                   self.__result,
                   self.sha256sum))

    __repr__ = __str__
